@@ -690,6 +690,22 @@ tag_changed (GtkTextTagTable            *table,
 	tag_table_changed (table, spell);
 }
 
+static void
+highlight_updated (GtkSourceBuffer            *buffer,
+                   GtkTextIter                *start,
+                   GtkTextIter                *end,
+                   GeditAutomaticSpellChecker *spell)
+{
+	check_range (spell, *start, *end, FALSE);
+}
+
+static void
+spell_tag_destroyed (GeditAutomaticSpellChecker *spell,
+                     GObject                    *where_the_object_was)
+{
+	spell->tag_highlight = NULL;
+}
+
 GeditAutomaticSpellChecker *
 gedit_automatic_spell_checker_new (GeditDocument     *doc,
 				   GeditSpellChecker *checker)
@@ -741,6 +757,11 @@ gedit_automatic_spell_checker_new (GeditDocument     *doc,
 			  G_CALLBACK (mark_set), 
 			  spell);
 
+	g_signal_connect (doc,
+	                  "highlight-updated",
+	                  G_CALLBACK (highlight_updated),
+	                  spell);
+
 	g_signal_connect (spell->spell_checker,
 			  "add_word_to_session",
 			  G_CALLBACK (add_word_signal_cb),
@@ -764,6 +785,10 @@ gedit_automatic_spell_checker_new (GeditDocument     *doc,
 				"underline", PANGO_UNDERLINE_ERROR,
 				NULL);
 
+	g_object_weak_ref (G_OBJECT (spell->tag_highlight),
+	                   (GWeakNotify)spell_tag_destroyed,
+	                   spell);
+
 	tag_table = gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (doc));
 
 	gtk_text_tag_set_priority (spell->tag_highlight, 
@@ -879,7 +904,7 @@ gedit_automatic_spell_checker_free_internal (GeditAutomaticSpellChecker *spell)
 
 	table = gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (spell->doc));
 
-	if (table != NULL)
+	if (table != NULL && spell->tag_highlight != NULL)
 	{
 		gtk_text_buffer_get_bounds (GTK_TEXT_BUFFER (spell->doc), 
 					    &start, 
